
/*
 * Declarations of two types of FVM routines:
 *  - GENERAL routines for arbitrary system of conservation laws
 *  - EQUATION SPECIFIC (Burgers', advection, MHD, Euler, Shallow Water, etc.) routines
 */

#ifndef EQUATION_H
#define EQUATION_H

#include "structures.h"

// === global variables

extern int NXl, NYl, NZl;
extern real dt, t, t_halt, halt_times[NSAVES+SAVE0], dx[NDIR];
extern volume<ConservedVars, 1> U;
extern volume<ConservedVars, 1> Flux[NDIR];
extern volume<ConservedVars, 1> Source;
extern volume<XtraVars, 1> Xtra;

// === global constants

const int ACTIVE_DIRS = (NX>1) + (NY>1) + (NZ>1);
const int WORK_FACTOR = 1 << ( ACTIVE_DIRS+1 - ORDER >= 0 ? ACTIVE_DIRS+1 - ORDER : 0 );

// === global functions

static int SIZE (int d) { return d==X?NXl:(d==Y?NYl:NZl); }
static bool HAVE_DIR (int d) { return (SIZE(d)>1); }

static int FLUX_NX (int d) { return d==X?NXl+1:NXl; }
static int FLUX_NY (int d) { return d==Y?NYl+1:NYl; }
static int FLUX_NZ (int d) { return d==Z?NZl+1:NZl; }

// === GENERAL FVM routines (defined in equation.cpp and elsewhere)

// prepare XtraVars and initialize GHOST cells
void step_init();

// compute needed quantities: mesh sizes, initial data, etc.
void model_init();

// used to compute initial data
void init_cell(PrimitiveVars &v, real x, real y, real z);

// Godunov-Powell source term
ConservedVars powell_calc (int i, int j, int k);
void powell_source (vector &us, vector &Bs, AllVariables &lV, AllVariables &rV, int i, int j, int k, int d);

// gravity interaction
ConservedVars gravity_calc (int i, int j, int k);

// applies boundary conditions
void set_ghost_x();
void set_ghost_y();
void set_ghost_z();
void bc_init();

// time integration routines
void time_step();
void sub_step (bool firsttime);

// === EQUATION SPECIFIC FVM routines (defined in corresponding equation_*.cpp)

// TODO: write comment
void calc_point_flux (ConservedVars &F, const AllVariables &V, int d);

// apply constraints to variables; for instance, pressure should be non-negative, etc.
void apply_constraints (ConservedVars &CV, XtraVars &XV);

// calculate timestep using CFL number
real calc_time_step();

// calculate XtraVars from ConservedVars
void calc_xtra (XtraVars &XU, const ConservedVars &U, int i, int j, int k);

// calculate AllVariables from PrimitiveVars
void calc_all_vars (AllVariables &V, const PrimitiveVars &P);

// calculate indicator for limiters
real indicator(int i, int j, int k, int d);

#endif
