
#ifndef NUMTYPES_H
#define NUMTYPES_H

#include <math.h>
#include <malloc.h>
#include <iostream>

#include "config.h"

static const int GX = GHOST_SIZE*(NX>1?1:0);
static const int GY = GHOST_SIZE*(NY>1?1:0);
static const int GZ = GHOST_SIZE*(NZ>1?1:0);

// define this for volume index checks.  NOT IMPLEMENTED
// #define INDEX_CHECKING

// dimension related constants.

static const int X = 0, Y = 1, Z = 2, NDIR = 3;
static const int XU = 3, YU = 4, ZU = 5;
static const int DX[] = { 1, 0, 0 };
static const int DY[] = { 0, 1, 0 };
static const int DZ[] = { 0, 0, 1 };

// define real so that we can easily compile single or double precision
// versions of the code.  All numbers are declared as real.

typedef double real;

// some useful inline numerical functions.

static inline real sqr(real n)
{
  return n*n;
}

static inline real sign(real n)
{
  if (n<=0) return -1; else return 1;
}

static inline real min(real a, real b)
{
  if (a < b) return a; else return b;
}

static inline real max(real a, real b)
{
  if (a > b) return a; else return b;
}

static inline real ssqrt(real x) {
  return sqrt(max(0, x));
}

// define pi

static const real pi = 3.1415926535897931;

// generic 3d vector type

template<class T>
class raw_vector {
  
  T v[NDIR];

public:
  
  raw_vector(T x, T y, T z) { v[X] = x; v[Y] = y; v[Z] = z; }
  raw_vector() { }
  
  const T &operator[](int i) const { return v[i]; }
  T &operator[](int i) { return v[i]; }
  
  T sqr() const { return v[X]*v[X] + v[Y]*v[Y] + v[Z]*v[Z]; }
  T norm() const { return fabs(v[X]) + fabs(v[Y]) + fabs(v[Z]); }
  
  void zero() {
    for (int i=0; i<NDIR; i++) v[i] = 0;
  }
  
  void operator+=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] += a.v[i];
  }
  
  void operator-=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] -= a.v[i];
  }
  
  void operator*=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] *= a.v[i];
  }
  
  void operator/=(const raw_vector<T> &a)
  {
    for (int i=0; i<NDIR; i++) v[i] /= a.v[i];
  }
  
  void operator*=(real n)
  {
    for (int i=0; i<NDIR; i++) v[i] *= n;
  }
  
  void operator/=(real n)
  {
    for (int i=0; i<NDIR; i++) v[i] /= n;
  }
  
  // needed for COMPUTE_KURTOSIS=1
  void operator-=(real n)
  {
    for (int i=0; i<NDIR; i++) v[i] -= n;
  }
  // end
  
  friend raw_vector<T> operator+(const raw_vector<T> &a, const raw_vector<T> &b)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i] + b.v[i];
    return r;
  }
  
  friend raw_vector<T> operator-(const raw_vector<T> &a, const raw_vector<T> &b)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i] - b.v[i];
    return r;
  }
  
  friend raw_vector<T> operator*(const raw_vector<T> &a, real n)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i]*n;
    return r;
  }
  
  friend raw_vector<T> operator*(real n, const raw_vector<T> &a)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i]*n;
    return r;
  }
  
  friend raw_vector<T> operator/(const raw_vector<T> &a, real n)
  { raw_vector<T> r;
    for (int i=0; i<NDIR; i++) r.v[i] = a.v[i]/n;
    return r;
  }
  
  // dot product
  friend real operator*(const raw_vector<T> &a, const raw_vector<T> &b)
  { real r=0;
    for (int i=0; i<NDIR; i++) r += a.v[i] * b.v[i];
    return r;
  }
};

// vector of reals
typedef raw_vector<real> vector;

// generic 3d volume array type (dynamic version)
// supports += and -= operators for hierarchical grids

template<class T, int GHOST, int VX0, int VY0, int VZ0>
class raw_volume {

private:
  
  int getLoc (int i, int j, int k) { return (i-VX0)*VNY*VNZ + (j-VY0)*VNZ + (k-VZ0); }
  
public:
  
  // pointer to actual array
  T* data;
  
  // grid sizes
  int NXl;
  int NYl;
  int NZl;
  
  // grid sizes including ghost cells
  int VNX;
  int VNY;
  int VNZ;
  
  void zero () {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++) {
          data[getLoc(i,j,k)] *= 0;
        }
  }
  
  // zero length allocations prevention
  raw_volume () : NXl(1), NYl(1), NZl(1), VNX(1 + 2*GX*GHOST), VNY(1 + 2*GY*GHOST), VNZ(1 + 2*GZ*GHOST) {
    data = (T*) malloc ( VNX * VNY * VNZ * sizeof(T) );
    if ( data == NULL ) {
      std::cout << " :: ERROR: out of memory." << std::endl;
      throw 1;
    }
    zero ();
  }
  
  raw_volume (int NXl_, int NYl_, int NZl_) : NXl(NXl_), NYl(NYl_), NZl(NZl_), VNX(NXl_ + 2*GX*GHOST), VNY(NYl_ + 2*GY*GHOST), VNZ(NZl_ + 2*GZ*GHOST) {
    data = (T*) malloc ( VNX * VNY * VNZ * sizeof(T) );
    if ( data == NULL ) {
      std::cout << " :: ERROR: out of memory." << std::endl;
      throw 1;
    }
    zero();
  }
  
  void reinit (int NXl_, int NYl_, int NZl_) {
    
    if ( NXl == NXl_ && NYl == NYl_ && NZl == NZl_ ) {
      zero();
      return;
    }
    
    NXl = NXl_;
    NYl = NYl_;
    NZl = NZl_;
    VNX = NXl_ + 2*GX*GHOST;
    VNY = NYl_ + 2*GY*GHOST;
    VNZ = NZl_ + 2*GZ*GHOST;
    T* data_ = (T*) realloc ( data, VNX * VNY * VNZ * sizeof(T) );
    if ( data_ == NULL ) {
      free (data);      // memory leak prevention
      std::cout << " :: ERROR: out of memory." << std::endl;
      throw 1;
    }
    data = data_;
    zero();
  }
  
  void reset () { zero(); }
  
  ~raw_volume () {
    free (data);        // memory leak prevention
  }
  
  // support for subscript operator [][][]
  struct slice_X {
    T* d;
    int VNZ;
    slice_X ( T *d_, int shift, int VNY_, int VNZ_ ) : d(&d_[shift*VNY_*VNZ_]), VNZ(VNZ_) {}
    
    struct slice_Y {
      T* d;
      slice_Y ( T *d_, int shift, int VNZ_ ) : d(&d_[shift*VNZ_]) {}
      
      T& operator[](const int k) const { return d[k]; }
    };
    
    slice_Y operator[](const int j) const { return slice_Y (d, j-VY0, VNZ); }
  };
  slice_X operator[](const int i) const { return slice_X (data, i-VX0, VNY, VNZ); }
  
  // validates compatibility of volumes for use with operators
  void validate (const raw_volume &a) const {
    if ( NXl != a.NXl || NYl != a.NYl || NZl != a.NZl ) {
      std::cout << " :: ERROR: non-compatible volume sizes." << std::endl;
      throw 1;
    }
  }
  
  void exp() {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++)
          Exp(data[getLoc(i,j,k)]);
  }
  
  void log() {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++)
          Log(data[getLoc(i,j,k)]);
  }
  
  void nonnegative() {
    for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++)
            NonNegative(data[getLoc(i,j,k)]);
  }
  
  void operator+=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] += a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] += a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }

  void operator-=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] -= a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] -= a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }
  
  void operator*=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] *= a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] *= a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }

  void operator/=(const raw_volume &a)
  {
    if ( NXl == a.NXl && NYl == a.NYl && NZl == a.NZl ) {
      for (int i=VX0; i<VNX+VX0; i++)
        for (int j=VY0; j<VNY+VY0; j++)
          for (int k=VZ0; k<VNZ+VZ0; k++) {
            data[getLoc(i,j,k)] /= a[i][j][k];
          }
    }
    
    else if ( NXl % a.NXl != 0 || NYl % a.NYl != 0 || NZl % a.NZl != 0 ) {
      std::cout << " :: ERROR: hierarchy of grids is violated." << std::endl;
      throw 1;
    }
    
    else {
      int factor_X = NXl / a.NXl;
      int factor_Y = NYl / a.NYl;
      int factor_Z = NZl / a.NZl;
      // ! increment non-ghost cells only !
      for (int i=0; i<NXl; i++)
        for (int j=0; j<NYl; j++)
          for (int k=0; k<NZl; k++) {
            data[getLoc(i,j,k)] /= a[i/factor_X][j/factor_Y][k/factor_Z];
          }
    }
  }
  
  void operator*=(const real r)
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
	for (int k=VZ0; k<VNZ+VZ0; k++) {
	  data[getLoc(i,j,k)] *= r;
	}
  }
  
  void operator/=(const real r)
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
	for (int k=VZ0; k<VNZ+VZ0; k++) {
	  data[getLoc(i,j,k)] /= r;
	}
  }
  
  void sqr ()
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++) {
          data[getLoc(i,j,k)] = data[getLoc(i,j,k)] * data[getLoc(i,j,k)];
    }
  }
  
  // === !!! operators creating NEW objects are very risky and can lead to memory inefficiency !!! be careful :)
  
  // needed for variance computation during parallel runs
  friend raw_volume operator-(const raw_volume &a, const raw_volume &b)
  {
    a.validate(b);
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] - b[i][j][k];
        }
    return v;
  }
  
  // needed for numerically stable variance computation
  friend raw_volume operator/(const raw_volume &a, const real r)
  {
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] / r;
        }
    return v;
  }
  
  // needed for COMPUTE_KURTOSIS=1
  
  friend raw_volume operator*(const raw_volume &a, const raw_volume &b)
  {
    a.validate(b);
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] * b[i][j][k];
        }
    return v;
  }
  
  friend raw_volume operator*(const raw_volume &a, const real r)
  {
    raw_volume v (a.NXl, a.NYl, a.NZl);
    for (int i=VX0; i<a.VNX+VX0; i++)
      for (int j=VY0; j<a.VNY+VY0; j++)
        for (int k=VZ0; k<a.VNZ+VZ0; k++) {
          v[i][j][k] = a[i][j][k] * r;
        }
    return v;
  }
  
  void operator-=(const real r)
  {
    for (int i=VX0; i<VNX+VX0; i++)
      for (int j=VY0; j<VNY+VY0; j++)
        for (int k=VZ0; k<VNZ+VZ0; k++) {
          data[getLoc(i,j,k)] -= r;
        }
  }
};

// 3d volume array type including ghost zones

template<class T, int GHOST>
class volume: public raw_volume<T, GHOST, -GX*GHOST, -GY*GHOST, -GZ*GHOST> {
public:
  volume () : raw_volume<T, GHOST, -GX*GHOST, -GY*GHOST, -GZ*GHOST> () {}
  volume (int NX_, int NY_, int NZ_) : raw_volume<T, GHOST, -GX*GHOST, -GY*GHOST, -GZ*GHOST> (NX_, NY_, NZ_) {}
};

#endif
