
#include "main.h"
#include "solver.h"
#include "ec_util.h"

///::title:: WB solver (Well-Balanced) | for Shallow Water (implemented for 1D only)
///::depends:: ec_util-sw
///::consts:: GHOST_SIZE=2
///::vars:: CFL=0.475 SONIC_EPS=0.1

void solver_calc_flux(ConservedVars &F, vector &us, vector &Bs,
                      const AllVariables &lV, const AllVariables &rV,
                      int i, int j, int k, int d)
{
  // === EC flux
  
  avg_flux (F, lV, rV, d);
  
  // === numerical diffusion
  
  // construction of matrix R
  
  real R [3][3] = { {1, 0, 1}, {0, 0, 0}, {0, 0, 0} };
  
  real sqrt_gh = sqrt ( GRAVITY * avg(lV.U.h, rV.U.h) );
  
  // 1st column of R
  R[1+d][0] = avg(lV.XU.u[d], rV.XU.u[d]) - sqrt_gh;
  R[2-d][0] = avg(lV.XU.u[1-d], rV.XU.u[1-d]);
  
  // 2nd column of R
  R[2-d][1] = sqrt_gh;
  
  // 3rd column of R
  R[1+d][2] = avg(lV.XU.u[d], rV.XU.u[d]) + sqrt_gh;
  R[2-d][2] = avg(lV.XU.u[1-d], rV.XU.u[1-d]);
  
  real scaling = sqrt(2*GRAVITY);
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      R[a][b] /= scaling;
  
  // construction of diagonal D
  
  real D [3];
  
  D[0] = fabs ( avg(lV.XU.u[d], rV.XU.u[d]) - sqrt_gh );
  D[1] = fabs ( avg(lV.XU.u[d], rV.XU.u[d])           );
  D[2] = fabs ( avg(lV.XU.u[d], rV.XU.u[d]) + sqrt_gh );
  
  // computing diffusion matrix RDR'
  
  real DR  [3][3] = {0.0};
  real RDR [3][3] = {0.0};
  
  // DR'
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      DR[a][b] += D[a] * R[b][a];
  
  // R(DR')
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      for (int c=0; c<3; c++)
        RDR[a][b] += R[a][c] * DR[c][b];
  
  // construction of change of variables (from equilibrium variables P to ConservedVars U) matrix U_P
  
  real l_a = max ( fabs( GRAVITY - sqr(lV.XU.u[0]) / lV.U.h ), SONIC_EPS );
  real r_a = max ( fabs( GRAVITY - sqr(rV.XU.u[0]) / rV.U.h ), SONIC_EPS );
  
  real U_P [3][3] = {0.0};
  
  U_P[0][0] = avg( 1/l_a, 1/r_a); 
  U_P[0][1] = - avg( lV.XU.u[0]/l_a, rV.XU.u[0]/r_a);
  U_P[1][1] = 1.0;
  
  // construction of vector jump in vector P
  
  real P [3] = {0.0};
  
  real llu = Xtra[i-2][j][k].u[0];
  real rru = Xtra[i+1][j][k].u[0];
  
  real llh = U[i-2][j][k].h;
  real rrh = U[i+1][j][k].h;
  
  P[0] = 0.5 * ( avg(rV.U.h, rrh) * avg(rV.XU.u[0], rru) - avg(lV.U.h, llh) * avg(lV.XU.u[0], llu) );
  P[1] = jump ( sqr(lV.XU.u[0])/2.0 + GRAVITY * ( lV.U.h + lV.XU.b ), sqr(rV.XU.u[0])/2.0 + GRAVITY * ( rV.U.h + rV.XU.b ) );
  
  // computing U_PP := U_P * P
  
  real U_PP [3] = {0.0};
  
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      U_PP[a] += U_P[a][b] * P[b];
  
  // computing diffusion term
  
  real RDRU_PP [3] = {0.0};
  
  for (int a=0; a<3; a++)
    for (int b=0; b<3; b++)
      RDRU_PP[a] += RDR[a][b] * U_PP[b];
  
  // adding diffusion to EC flux
  
  F.h     -= 0.5 * RDRU_PP[0];
  F.hu[0] -= 0.5 * RDRU_PP[1];
  F.hu[1] -= 0.5 * RDRU_PP[2];
}
